use std::collections::VecDeque;
use std::net::SocketAddr;
use std::rc::Rc;
use std::cell::{Cell,RefCell};
use std::ops::DerefMut;

use mio::net::UdpSocket;
use mio::Token;
use mio::Interest;

use crate::Fleck;
use crate::prelude::*;
use crate::service::event::NoPriorityTag;
use super::AbstractFdService;

pub trait UdpData: 'static + Into<(SocketAddr, Vec<u8>)> + From<(SocketAddr, Vec<u8>)> {}

impl<T: 'static + Into<(SocketAddr, Vec<u8>)> + From<(SocketAddr, Vec<u8>)>> UdpData for T { }

pub struct UdpService<Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, Outgoing: ChannelSpec<Data = DataType>, DataType: UdpData> {
    pub(super) socket: RefCell<UdpSocket>,
    pub(super) token: Token,
    pub(super) poll: Rc<RefCell<mio::Poll>>,
    pub(super) send_queue: RefCell<VecDeque<(SocketAddr, Vec<u8>)>>,
    pub(super) write_registered: Cell<bool>,
    pub(super) _ghost: std::marker::PhantomData<(Incoming, Outgoing, DataType)>,
}

impl<Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, DataType: UdpData> Service for Rc<UdpService<Incoming, Outgoing, DataType>> {
    fn register_channels(&self, eroot: &mut crate::prelude::EventRoot) { 
        eroot.channel::<Outgoing>().sub_opt(self, UdpService::handle_outgoing);
    }
}

impl<Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, DataType: UdpData> UdpService<Incoming, Outgoing, DataType> {
    fn clear_write_interest(&self) {
        self.poll
            .borrow()
            .registry()
            .reregister(
                self.socket.borrow_mut().deref_mut(),
                self.token,
                Interest::READABLE,
            )
            .expect("couldn't update interest?");
        self.write_registered.set(false);
    }

    fn register_write_interest(&self) {
        if self.write_registered.get() { return }
        self.poll
            .borrow()
            .registry()
            .reregister(
                self.socket.borrow_mut().deref_mut(),
                self.token,
                Interest::READABLE.add(Interest::WRITABLE),
            )
            .expect("couldn't update interest?");
        self.write_registered.set(true);
    }


    fn handle_outgoing(&self, _api: &Fleck, packet: DataType) -> Option<DataType> {
        self.send_queue.borrow_mut().push_back(packet.into());

        self.register_write_interest();

        None
    }
}

impl<Incoming: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, Outgoing: ChannelSpec<Priority = NoPriorityTag, Data = DataType>, DataType: UdpData> AbstractFdService for UdpService<Incoming, Outgoing, DataType> {
    fn ready_read(&self, api: &Fleck) {
        let mut buf = [0u8; 2048];
        while let Ok((len, addr)) = self.socket.borrow_mut().recv_from(&mut buf) {
            api.queue::<Incoming>((addr, Vec::from(&buf[..len])).into());
        }
    }

    fn ready_write(&self, _api: &Fleck) {
        let mut queue = self.send_queue.borrow_mut();
        while !queue.is_empty() {
            let pkt = queue.front().unwrap();
            match self.socket.borrow_mut().send_to(&pkt.1, pkt.0) {
                Ok(len) => {
                    if len != pkt.1.len() {
                        log::warn!(
                            "Packet of {} bytes truncated to {} bytes when sending",
                            pkt.1.len(),
                            len
                        );
                    }
                    queue.pop_front();
                }
                Err(e) => {
                    log::info!("Error while sending packet: {:?}", e);
                    break;
                }
            }
        }

        if queue.is_empty() { self.clear_write_interest() }
    }

    fn token(&self) -> Token {
        self.token
    }
}
